from .AISquaredAPIException import AISquaredAPIException
from .additional_utils import _check_results_length
from .NoResultsFoundError import NoResultsFoundError
from aisquared.base import ENDPOINTS
import pandas as pd
import requests


def _list_model_feedback(
        url: str,
        headers: dict,
        model_id: str,
        limit: int,
        as_df: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    List feedback from a model

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    model_id : str
        The ID of the model
    limit : int
        The maximum number of feedback items to return
    as_df : bool
        Whether to return the data as a Pandas DataFrame
    """

    url = f'{url}/{ENDPOINTS["feedback"]}/models?modelId={model_id}&page=1&pageSize={limit}'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        if resp.status_code == 404:
            raise NoResultsFoundError('No Results Found')
        raise AISquaredAPIException(resp.json())

    if as_df:
        df = pd.DataFrame(resp.json()['data']['modelFeedback'])
        _check_results_length(df)
        return df
    return resp.json()['data']


def _list_prediction_feedback(
        url: str,
        headers: dict,
        prediction_id: str,
        as_df: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    List prediction feedback for a prediction ID

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    prediction_id : str
        The prediction ID
    as_df : bool
        Whether to return the data as a Pandas DataFrame
    """

    url = f'{url}/{ENDPOINTS["feedback"]}/predictions?modelId={prediction_id}'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    if as_df:
        df = pd.DataFrame(resp.json()['data'])
        _check_results_length(df)
        return df
    return resp.json()['data']


def _list_model_prediction_feedback(
        url: int,
        headers: dict,
        model_id: str,
        as_df: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    List prediction feedback from a model

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    model_id : str
        The ID of the model
    as_df : bool
        Whether to return the data as a Pandas DataFrame
    """

    url = f'{url}/{ENDPOINTS["feedback"]}/predictions?modelId={model_id}'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    if as_df:
        data = resp.json()['data']
        df = pd.concat([pd.json_normalize(v)
                       for v in data.values()]).reset_index(drop=True)
        _check_results_length(df)
        return df

    return resp.json()



from typing import Union

from .DatabricksAPIException import DatabricksAPIException
from aisquared.base import DIRECTORY

from getpass import getpass
import pandas as pd
import warnings
import requests
import base64
import json
import os

CLIENT_CONFIG_FILE = os.path.join(DIRECTORY, '.databricks.json')


class DatabricksClient:
    """
    Client for working with a connected Databricks environment

    When using the client for the first time, it is important to authenticate the
    client using the `client.login()` method. When doing so, the client will ask for any
    required information interactively.

    >>> import aisquared
    >>> client = aisquared.platform.DatabricksClient()
    >>> # If you have never logged in before, run the following code:
    >>> client.login()
    >>> # Interactive session requesting required information
    """

    def __init__(
            self
    ):
        """
        Initialize the object
        """
        try:
            self._load_info(CLIENT_CONFIG_FILE)
        except Exception:
            warnings.warn(
                'It appears you are not authenticated to Databricks yet. Please run Client.login() before performing any action'
            )

    def login(
            self,
            url: str = None,
            username: str = None,
            token: str = None,
            persist: bool = True
    ) -> None:
        """
        Log in to the Databricks environment programmatically

        >>> import aisquared
        >>> client = aisquared.platform.DatabricksClient()
        >>> client.login()
        Enter URL: {Databricks_workspace_url}
        Enter Username: your.email@your_domain.com
        Enter Secret Token: <hidden>

        Parameters
        ----------
        url : str or None (default None)
            The URL of the Databricks workspace
        username : str or None (default None)
            The username in the Databricks workspace
        token : str or None (default None)
            The secret token for the Databricks workspace
        persist : bool (default True)
            Whether to persist the login information, eliminating
            the need to run this command again in the future
        """
        if url is None:
            url = input('Enter URL: ')
        if username is None:
            username = input('Enter Username: ')
        if token is None:
            token = getpass('Enter Secret Token: ')

        if persist:
            with open(CLIENT_CONFIG_FILE, 'w') as f:
                json.dump(
                    {
                        'url': url,
                        'username': username,
                        'token': token
                    },
                    f
                )

            self._load_info()

        else:
            self._base_url = url
            self._username = username
            self._token = token

    def _load_info(self, config_file: str = CLIENT_CONFIG_FILE) -> None:
        """
        NOT MEANT TO BE CALLED BY THE END USER

        Load login information
        """
        with open(config_file, 'r') as f:
            data = json.load(f)

        self._base_url = data['url']
        self._username = data['username']
        self._token = data['token']

    @property
    def headers(self) -> dict:
        """API headers for calls to the API"""
        return {
            'authorization': f'Bearer {self._token}',
            'Content-Type': 'application/json',
            'User-Agent': 'AI Squared/1.0'
        }

    @property
    def username(self) -> str:
        """The user's username"""
        return self._username

    @property
    def base_url(self) -> str:
        """The base URL for the workspace"""
        return self._base_url

    @property
    def token(self) -> str:
        """The token to use for the workspace"""
        return '*' * len(self._token)

    def list_workspace(self, as_df: bool = True) -> Union[pd.DataFrame, dict]:
        """
        List files in the connected Databricks workspace

        Parameters
        ----------
        as_df : bool (default True)
            Whether to return the results as a pandas DataFrame

        Returns
        -------
        results : dict or pd.DataFrame
            The files in the workspace
        """
        with requests.Session() as sess:
            resp = sess.get(
                url=f'{self.base_url}/api/2.0/workspace/list',
                headers=self.headers,
                json={'path': f'/Users/{self.username}'}
            )
        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        if as_df:
            return pd.json_normalize(resp.json()['objects'])
        return resp.json()

    def upload_to_workspace(self, filename: str, overwrite: bool = False) -> bool:
        """
        Upload a file to the workspace

        Parameters
        ----------
        filename : str
            The name of the file to upload
        overwrite : bool (default False)
            Whether to overwrite the file if one of the same name already
            exists in the workspace

        Returns
        -------
        success : bool
            Whether the upload was successful
        """
        file_b64 = base64.b64encode(
            open(filename, 'rb').read()).decode('ascii')
        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.0/workspace/import',
                headers=self.headers,
                json={
                    'path': f'/Users/{self.username}/{os.path.basename(filename)}',
                    'language': 'PYTHON',
                    'overwrite': overwrite,
                    'content': file_b64
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.ok

    def download_from_workspace(self, filename: str) -> str:
        """
        Download a file from the workspace

        Parameters
        ----------
        filename : str
            The filename of the file to download

        Returns
        -------
        contents : str
            The contents of the file
        """
        if not filename.startswith(f'/Users/{self.username}/'):
            filename = f'/Users/{self.username}/{filename}'
        with requests.Session() as sess:
            resp = sess.get(
                url=f'{self.base_url}/api/2.0/workspace/export',
                headers=self.headers,
                json={
                    'path': filename,
                    'format': 'SOURCE',
                    'direct_download': True
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.text

    def delete_from_workspace(self, filename: str) -> bool:
        """
        Delete a file from the workspace

        Parameters
        ----------
        filename : str
            The name of the file to delete

        Returns
        -------
        success : bool
            Whether the operation is successful
        """
        if not filename.startswith(f'/Users/{self.username}/'):
            filename = f'/Users/{self.username}/{filename}'
        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.0/workspace/delete',
                headers=self.headers,
                json={
                    'path': filename
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.ok

    def create_job(
            self,
            job_name: str,
            tasks: list,
            libraries: list,
            compute_name: str,
            spark_version: str,
            node_type_id: str,
            cron_syntax: str = None,
            timezone: str = None
    ) -> bool:
        """
        Create a job using notebooks and/or scripts in the workspace

        Parameters
        ----------
        job_name : str
            The name for the job
        tasks : list of dict
            List of {task_name : task_script} dictionary pairs to run in the job
        libraries : list of str
            The dependent libraries to install on all compute
        compute_name : str
            The name of the compute to provision specifically for this job
        spark_version : str
            The version of Spark to use on the compute instances
        node_type_id : str
            The node type to use
        cron_syntax : str or None (default None)
            If the job is to be set to a schedule, the cron syntax for that schedule
        timezone : str or None (default None)
            The timezone to set the schedule to, if cron syntax is provided

        Returns
        -------
        success : bool
            Whether the create job call was successful
        """

        # Create the array of libraries
        library_array = [
            {'pypi': {'package': library}} for library in libraries
        ]

        # Create the array of tasks from the name : notebook locations in the dictionary of tasks
        task_array = []
        for i in range(len(tasks)):
            task_name = list(tasks[i].keys())[0]
            task_notebook = list(tasks[i].values())[0]

            task_dict = {
                'task_key': task_name,
                'run_if': 'ALL_SUCCESS',
                'notebook_task': {
                    'notebook_path': task_notebook,
                    'source': 'WORKSPACE'
                },
                'job_cluster_key': compute_name,
                'libraries': library_array,
                'timeout_seconds': 0,
                'email_notifications': {},
                'notification_settings': {
                    'no_alert_for_skipped_runs': False,
                    'no_alert_for_canceled_runs': False,
                    'alert_on_last_attempt': False
                },
                'webhook_notifications': {},
            }

            if i != 0:
                task_dict['depends_on'] = [
                    {
                        'task_key': list(tasks[i - 1].keys())[0]
                    }
                ]

            task_array.append(task_dict)

        # Create the job_clusters dictionary
        job_clusters = [{
            'job_cluster_key': compute_name,
            'new_cluster': {
                'cluster_name': '',
                'spark_version': spark_version,
                'spark_conf': {
                    'spark.master': 'local[*, 4]'
                },
                'num_workers': 0,
                'node_type_id': node_type_id,
                'enable_elastic_disk': True,
                'data_security_mode': 'LEGACY_SINGLE_USER_STANDARD',
                'runtime_engine': 'STANDARD',
                'spark_env_vars': {
                    'PYSPARK_PYTHON': '/databricks/python3/bin/python'
                }
            }
        }]

        # Create the entire json to be sent with the request
        job_dict = {
            'name': job_name,
            'email_notifications': {
                'no_alert_for_skipped_runs': False
            },
            'webhook_notifications': {},
            'timeout_seconds': 0,
            'max_concurrent_runs': 1,
            'tasks': task_array,
            'job_clusters': job_clusters,
            'run_as': {
                'user_name': self.username
            }
        }

        if cron_syntax and timezone:
            job_dict['schedule'] = {
                'quartz_cron_expression': cron_syntax,
                'timezone_id': timezone,
                'pause_status': 'UNPAUSED'
            }

        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.1/jobs/create',
                headers=self.headers,
                json=job_dict
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.ok

    def list_jobs(self, as_df: bool = True) -> Union[dict, pd.DataFrame]:
        """
        List all jobs in the workspace

        Parameters
        ----------
        as_df : bool (default True)
            Whether to return a pandas DataFrame

        Returns
        -------
        jobs : dict or pandas DataFrame
            The jobs that exist in the workspace
        """
        with requests.Session() as sess:
            resp = sess.get(
                url=f'{self.base_url}/api/2.1/jobs/list',
                headers=self.headers,
                json={
                    'limit': 100
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        if as_df:
            return pd.json_normalize(resp.json()['jobs'])

        return resp.json()

    def delete_job(self, job_id: str) -> bool:
        """
        Delete a job from the workspace

        Parameters
        ----------
        job_id : str
            The ID of the job to delete

        Returns
        -------
        success : bool
            Whether the delete operation was successful
        """
        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.1/jobs/delete',
                headers=self.headers,
                json={
                    'job_id': job_id
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.ok

    def run_job(self, job_id: str) -> int:
        """
        Run a job

        Parameters
        ----------
        job_id : str
            The ID of the job to run

        Returns
        -------
        run_id : int
            The ID of the specific run that was created
        """
        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.1/jobs/run-now',
                headers=self.headers,
                json={
                    'job_id': job_id
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.json()['run_id']

    def update_job(
            self,
            job_id: int,
            job_name: str,
            tasks: list,
            libraries: list,
            compute_name: str,
            spark_version: str,
            node_type_id: str,
            cron_syntax: str = None,
            timezone: str = None
    ) -> bool:
        """
        Update a job by Job ID using notebooks and/or scripts in the workspace

        Parameters
        ----------
        job_id : int
            The unique identifier of the job to update
        job_name : str
            The new name for the job
        tasks : list of dict
            List of {task_name : task_script} dictionary pairs to run in the updated job
        libraries : list of str
            The dependent libraries to install on all compute for the new job
        compute_name : str
            The name of the compute to provision specifically for the new job
        spark_version : str
            The version of Spark to use on the compute instances
        node_type_id : str
            The node type to use
        cron_syntax : str or None (default None)
            If the new job is to be set to a schedule, the cron syntax for that schedule
        timezone : str or None (default None)
            The timezone to set the schedule to, if cron syntax is provided

        Returns
        -------
        success : bool
            Whether the update job call was successful
        """
        # Create the array of libraries
        library_array = [
            {'pypi': {'package': library}} for library in libraries
        ]

        # Create the array of tasks from the name : notebook locations in the dictionary of tasks
        task_array = []
        for i in range(len(tasks)):
            task_name = list(tasks[i].keys())[0]
            task_notebook = list(tasks[i].values())[0]

            task_dict = {
                'task_key': task_name,
                'run_if': 'ALL_SUCCESS',
                'notebook_task': {
                    'notebook_path': task_notebook,
                    'source': 'WORKSPACE'
                },
                'job_cluster_key': compute_name,
                'libraries': library_array,
                'timeout_seconds': 0,
                'email_notifications': {},
                'notification_settings': {
                    'no_alert_for_skipped_runs': False,
                    'no_alert_for_canceled_runs': False,
                    'alert_on_last_attempt': False
                },
                'webhook_notifications': {},
            }

            if i != 0:
                task_dict['depends_on'] = [
                    {
                        'task_key': list(tasks[i - 1].keys())[0]
                    }
                ]

            task_array.append(task_dict)

        # Create the job_clusters dictionary
        job_clusters = [{
            'job_cluster_key': compute_name,
            'new_cluster': {
                'cluster_name': '',
                'spark_version': spark_version,
                'spark_conf': {
                    'spark.master': 'local[*, 4]'
                },
                'num_workers': 0,
                'node_type_id': node_type_id,
                'enable_elastic_disk': True,
                'data_security_mode': 'LEGACY_SINGLE_USER_STANDARD',
                'runtime_engine': 'STANDARD',
                'spark_env_vars': {
                    'PYSPARK_PYTHON': '/databricks/python3/bin/python'
                }
            }
        }]

        # Create the entire json to be sent with the request
        job_dict = {
            'name': job_name,
            'email_notifications': {
                'no_alert_for_skipped_runs': False
            },
            'webhook_notifications': {},
            'timeout_seconds': 0,
            'max_concurrent_runs': 1,
            'tasks': task_array,
            'job_clusters': job_clusters,
            'run_as': {
                'user_name': self.username
            }
        }

        if cron_syntax and timezone:
            job_dict['schedule'] = {
                'quartz_cron_expression': cron_syntax,
                'timezone_id': timezone,
                'pause_status': 'UNPAUSED'
            }

        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.1/jobs/reset',
                headers=self.headers,
                json={
                    'job_id': job_id,
                    'new_settings': job_dict
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.ok

    def list_served_models(self, as_df: bool = True) -> Union[dict, pd.DataFrame]:
        """
        List served models in the workspace

        Parameters
        ----------
        as_df : bool (default True)
            Whether to return results as a pandas DataFrame

        Returns
        -------
        models : dict or pandas DataFrame
            The models served in the workspace
        """
        with requests.Session() as sess:
            resp = sess.get(
                url=f'{self.base_url}/api/2.0/serving-endpoints',
                headers=self.headers
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        if as_df:
            return pd.json_normalize(resp.json()['endpoints'])
        else:
            return resp.json()

    def delete_served_model(self, model_name: str) -> bool:
        """
        Delete a served model in the workspace

        Parameters
        ----------
        model_name : str
            The name of the model to delete

        Returns
        -------
        success : bool
            Whether the delete operation was successful
        """
        with requests.Session() as sess:
            resp = sess.delete(
                url=f'{self.base_url}/api/2.0/serving-endpoints/{model_name}',
                headers=self.headers
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.ok

    def create_served_model(
            self,
            model_name: str,
            model_version: str,
            workload_size: str,
            scale_to_zero_enabled: bool = True,
            workload_type: str = 'CPU'
    ) -> dict:
        """
        Create a model serving endpoint

        Parameters
        ----------
        model_name : str
            The name of the model to serve
        model_version : str
            The version of the model to serve
        workload_size : str
            The workload size of the serving endpoint
        scale_to_zero_enabled : bool (default True)
            Whether to allow for scaling the endpoint to zero
        workload type : str (default 'CPU')
            The workload type - either 'CPU' or 'GPU'

        Returns
        -------
        configuration : dict
            Configuration information about the serving endpoint
        """
        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.0/serving-endpoints',
                headers=self.headers,
                json={
                    'name': model_name,
                    'config': {
                        'served_models': [{
                            'model_name': model_name,
                            'model_version': model_version,
                            'workload_size': workload_size,
                            'scale_to_zero_enabled': scale_to_zero_enabled,
                            'workload_type': workload_type
                        }]
                    }
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.json()

    def list_compute(self, as_df: bool = True) -> Union[dict, pd.DataFrame]:
        """
        List compute in the workspace

        Parameters
        ----------
        as_df : bool (default True)
            Whether to return a pandas DataFrame

        Returns
        -------
        compute : dict or pd.DataFrame
            The compute resources in the workspace
        """
        with requests.Session() as sess:
            resp = sess.get(
                url=f'{self.base_url}/api/2.0/clusters/list',
                headers=self.headers
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        if as_df:
            return pd.json_normalize(resp.json()['clusters'])

        return resp.json()

    def delete_compute(self, compute_id: str) -> bool:
        """
        Delete a compute resource in the workspace

        Parameters
        ----------
        compute_id : str
            The ID for the compute to delete

        Returns
        -------
        success : bool
            Whether the operation was successful
        """
        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.0/clusters/permanent-delete',
                headers=self.headers,
                json={
                    'cluster_id': compute_id
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.ok

    def create_compute(
            self,
            compute_name: str,
            spark_version: str,
            node_type_id: str
    ) -> dict:
        """
        Create a compute resource

        Parameters
        ----------
        compute_name : str
            The name of the compute to create
        spark_version : str
            The spark version to use for the compute resource
        node_type_id : str
            The node type ID to use

        Returns
        -------
        compute_info : dict
            The information about the created compute resource
        """
        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.0/clusters/create',
                headers=self.headers,
                json={
                    "cluster_name": compute_name,
                    "spark_version": spark_version,
                    "node_type_id": node_type_id,
                    "num_workers": 0,
                    "spark_conf": {
                        "spark.databricks.cluster.profile": "singleNode",
                        "spark.master": "[*, 4]"
                    },
                    "custom_tags": {
                        "ResourceClass": "SingleNode"
                    }
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.json()

    def start_compute(self, compute_id: str) -> bool:
        """
        Start a compute resource

        Parameters
        ----------
        compute_id : str
            The ID of the compute to start

        Returns
        -------
        success : bool
            Whether the start operation was successful
        """
        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.0/clusters/start',
                headers=self.headers,
                json={
                    'cluster_id': compute_id
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.ok

    def stop_compute(self, compute_id: str) -> bool:
        """
        Stop a compute resource

        Parameters
        ----------
        compute_id : str
            The ID of the compute to start

        Returns
        -------
        success : bool
            Whether the stop operation was successful
        """
        with requests.Session() as sess:
            resp = sess.post(
                url=f'{self.base_url}/api/2.0/clusters/delete',
                headers=self.headers,
                json={
                    'cluster_id': compute_id
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.ok

    def list_registered_models(self, as_df: bool = True) -> Union[dict, pd.DataFrame]:
        """
        List registered models in the workspace

        Parameters
        ----------
        as_df : bool (default True)
            Whether to return a pandas DataFrame

        Returns
        -------
        models : dict or pandas DataFrame
            The models in the workspace
        """
        with requests.Session() as sess:
            resp = sess.get(
                url=f'{self.base_url}/api/2.0/mlflow/registered-models/list',
                headers=self.headers,
                json={
                    'max_results': 1000
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        if as_df:
            return pd.json_normalize(resp.json()['registered_models'])

        return resp.json()

    def delete_registered_model(self, model_name: str) -> bool:
        """
        Delete a registered model

        Parameters
        ----------
        model_name : str
            The name of the model to delete

        Returns
        -------
        success : bool
            Whether the delete operation was successful
        """
        with requests.Session() as sess:
            resp = sess.delete(
                url=f'{self.base_url}/api/2.0/mlflow/registered-models/delete',
                headers=self.headers,
                json={
                    'name': model_name
                }
            )

        if not resp.ok:
            raise DatabricksAPIException(resp.text)

        return resp.ok



from typing import Union
from aisquared.base import BaseObject, ALLOWED_HOWS


class TextHarvester(BaseObject):
    """
    Object to harvest text

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.harvesting.TextHarvester(
        how = 'all',
        body_only = True
    )
    >>> my_obj.to_dict()
    {'className': 'TextHarvester',
    'params': {'how': 'all',
    'regex': None,
    'flags': 'gu',
    'bodyOnly': True,
    'limit': None}}

    """

    def __init__(
        self,
        how: str = ALLOWED_HOWS[0],
        regex: str = None,
        flags: str = 'gu',
        body_only: bool = False,
        keywords: Union[str, list] = None,
        limit: int = None
    ):
        """
        Parameters
        ----------
        how : str (default 'all')
            How to harvest text (supports ['all', 'regex', 'keywords'])
        regex : str or None (default None)
            Javascript-compatible regular expression to use to harvest individual strings
        flags : str or None (default 'gu')
            Flags to use with the Regex
        body_only : bool (default False)
            Whether to only harvest text from the body of the webpage
        keywords : list of str or None (default None)
            If provided, keywords to search for
        limit : int or None (default None)
            The limit to the number of times to perform the harvesting, if provided
        """
        super().__init__()
        self.how = how
        self.regex = regex
        self.flags = flags
        self.body_only = body_only
        self.keywords = keywords
        self.limit = limit

    @property
    def how(self):
        return self._how

    @how.setter
    def how(self, value):
        if value not in ALLOWED_HOWS:
            raise ValueError(
                f"how must be one of {ALLOWED_HOWS}, got {value}")
        self._how = value

    @property
    def regex(self):
        return self._regex

    @regex.setter
    def regex(self, value):
        self._regex = str(value) if value is not None else value

    @property
    def flags(self):
        return self._flags

    @flags.setter
    def flags(self, value):
        self._flags = value

    @property
    def body_only(self):
        return self._body_only

    @body_only.setter
    def body_only(self, value):
        if not isinstance(value, bool):
            raise TypeError('body_only must be Boolean')
        self._body_only = value

    @property
    def limit(self):
        return self._limit

    @limit.setter
    def limit(self, value):
        if not isinstance(value, int) and value is not None:
            raise TypeError('limit must be int or None')
        self._limit = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        if self.how == 'keywords':
            return {
                'className': 'TextHarvester',
                'params': {
                    'how': 'regex',
                    'regex': '|'.join(self.keywords),
                    'flags': self.flags,
                    'bodyOnly': self.body_only,
                    'limit': self.limit
                }
            }

        return {
            'className': 'TextHarvester',
            'params': {
                'how': self.how,
                'regex': self.regex,
                'flags': self.flags,
                'bodyOnly': self.body_only,
                'limit': self.limit
            }
        }



from typing import Union
from aisquared.base import BaseObject


class DocumentRendering(BaseObject):
    """
    Object which dictates how to render predictions on entire documents

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.DocumentRendering()
    >>> my_obj.to_dict()
    {'className': 'DocumentRendering',
    'params': {'predictionKey': 'className',
    'words': None,
    'documents': None,
    'includeProbability': False,
    'probabilityKey': 'probability',
    'underlineColor': 'blue',
    'classes': None,
    'thresholdKey': None,
    'thresholdValue': None}}

    """

    def __init__(
        self,
        prediction_key: str = 'className',
        words: Union[list, dict, str] = None,
        documents: Union[list, dict, str] = None,
        include_probability: bool = False,
        probability_key: str = 'probability',
        underline_color: str = 'blue',
        classes: list = None,
        threshold_key: str = None,
        threshold_value: Union[int, float] = None
    ):
        """
        Parameters
        ----------
        prediction_key : str (default 'className')
            The key from the postprocessing objects to use to get the prediction
            For classification cases, use 'className'. For regression cases, use 'value'
        words : list, dict, string, or None (default None)
            Words to be highlighted
        documents : list, dict, string, or None (default None)
            Documents (typically in the form of URLs) to be recommended
        include_probability : bool (default False)
            Whether to include probabilities
        probability_key : str (default 'probability')
            Key to use to retrieve probability from results
        underline_color : str (default 'blue')
        classes : None or list (default None)
            If provided, list of classes that will be rendered
        threshold_key : None or str (default None)
            If provided, the key to threshold on
        threshold_value : None or numeric (default None)
            If provided with threshold_key, the value to threshold rendering upon
        """
        super().__init__()
        self.prediction_key = prediction_key
        self.words = words
        self.documents = documents
        self.include_probability = include_probability
        self.probability_key = probability_key
        self.underline_color = underline_color
        self.classes = classes
        self.threshold_key = threshold_key
        self.threshold_value = threshold_value

    @property
    def prediction_key(self):
        return self._prediction_key

    @prediction_key.setter
    def prediction_key(self, value):
        self._prediction_key = value

    @property
    def words(self):
        return self._words

    @words.setter
    def words(self, value):
        self._words = value

    @property
    def documents(self):
        return self._documents

    @documents.setter
    def documents(self, value):
        self._documents = value

    @property
    def include_probability(self):
        return self._include_probability

    @include_probability.setter
    def include_probability(self, value):
        self._include_probability = value

    @property
    def probability_key(self):
        return self._probability_key

    @probability_key.setter
    def probability_key(self, value):
        self._probability_key = value

    @property
    def underline_color(self):
        return self._underline_color

    @underline_color.setter
    def underline_color(self, value):
        self._underline_color = value

    @property
    def classes(self):
        return self._classes

    @classes.setter
    def classes(self, value):
        self._classes = value

    @property
    def threshold_key(self):
        return self._threshold_key

    @threshold_key.setter
    def threshold_key(self, value):
        self._threshold_key = value

    @property
    def threshold_value(self):
        return self._threshold_value

    @threshold_value.setter
    def threshold_value(self, value):
        self._threshold_value = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'DocumentRendering',
            'params': {
                'predictionKey': self.prediction_key,
                'words': self.words,
                'documents': self.documents,
                'includeProbability': self.include_probability,
                'probabilityKey': self.probability_key,
                'underlineColor': self.underline_color,
                'classes': self.classes,
                'thresholdKey': self.threshold_key,
                'thresholdValue': self.threshold_value
            }
        }



"""
Additional utilities to use with the `aisquared` package. These utilities currently consist of two functions, the
`mimic_model` and `get_model` functions. They utilize functionality that exists in our open source package BeyondML
to train teacher-student models

To see in-depth examples of how to use these functions, please visit our GitHub repository at https://github.com/AISquaredInc/MimicModelExamples
"""

from .utils import mimic_model, get_model



"""
Utilities for interacting with the AI Squared Platform.

The primary class within this subpackage is the `AISquaredPlatformClient` class, which has the
capabilities to interact with much of the functionality in the AI Squared platform. For more
information about this class, please see its documentation.
"""

from .AISquaredPlatformClient import AISquaredPlatformClient
from .DatabricksClient import DatabricksClient



from typing import Union
from aisquared.base import BaseObject, ALLOWED_PADS


class Tokenize(BaseObject):
    """Preprocessing Step to tokenize text

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.text.TextPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.text.Tokenize()
    )

    """

    def __init__(
        self,
        split_sentences: bool = False,
        split_words: bool = True,
        token_pattern: str = '\b\w\w+\b'
    ):
        """
        Parameters
        ----------
        split_sentences : bool (default False)
            Whether to split on sentences first
        split_words : bool (default True)
            Whether to split on words
        token_pattern : str (default '(?u)\\b\\w\\w+\\b')
            Regex to tokenize on
        """
        super().__init__()
        self.split_sentences = split_sentences
        self.split_words = split_words
        self.token_pattern = token_pattern

    @property
    def split_sentences(self):
        return self._split_sentences

    @split_sentences.setter
    def split_sentences(self, value):
        if not isinstance(value, bool):
            raise TypeError('split_sentences must be bool')
        self._split_sentences = value

    @property
    def split_words(self):
        return self._split_words

    @split_words.setter
    def split_words(self, value):
        if not isinstance(value, bool):
            raise TypeError('split_words must be bool')
        self._split_words = value

    @property
    def token_pattern(self):
        return self._token_pattern

    @token_pattern.setter
    def token_pattern(self, value):
        if not isinstance(value, str):
            raise TypeError('token_pattern must be string')
        self._token_pattern = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'Tokenize',
            'params': {
                'splitSentences': self.split_sentences,
                'splitWords': self.split_words,
                'tokenPattern': self.token_pattern
            }
        }


class RemoveCharacters(BaseObject):
    """Preprocessing step to remove characters from text

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.text.TextPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.text.RemoveCharacters()
    )
    """

    def __init__(
        self,
        remove_digits: bool = True,
        remove_punctuation: bool = True
    ):
        """
        Parameters
        ----------
        remove_digits : bool (default True)
            Whether to remove digits from input text
        remove_punctuation : bool (default True)
            Whether to remove punctuation from input text
        """
        super().__init__()
        self.remove_digits = remove_digits
        self.remove_punctuation = remove_punctuation

    @property
    def remove_digits(self):
        return self._remove_digits

    @remove_digits.setter
    def remove_digits(self, value):
        if not isinstance(value, bool):
            raise TypeError('remove_digits must be bool')
        self._remove_digits = value

    @property
    def remove_punctuation(self):
        return self._remove_punctuation

    @remove_punctuation.setter
    def remove_punctuation(self, value):
        if not isinstance(value, bool):
            raise TypeError('remove_punctuation must be bool')
        self._remove_punctuation = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'RemoveCharacters',
            'params': {
                'removeDigits': self.remove_digits,
                'removePunctuation': self.remove_punctuation
            }
        }


class ConvertToCase(BaseObject):
    """Text preprocessing object to convert inputs to all lowercase or all uppercase

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.text.TextPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.text.ConvertToCase()
    )
    """

    def __init__(
        self,
        lowercase: bool = True
    ):
        """
        Parameters
        ----------
        lowercase : bool (default True)
            Whether to convert to lower case. If False, converts to all uppercase
        """
        super().__init__()
        self.lowercase = lowercase

    @property
    def lowercase(self):
        return self._lowercase

    @lowercase.setter
    def lowercase(self, value):
        if not isinstance(value, bool):
            raise TypeError('lowercase must be bool')
        self._lowercase = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ConvertToCase',
            'params': {
                'lowercase': self.lowercase
            }
        }


class ConvertToVocabulary(BaseObject):
    """Text preprocessing object to convert tokens to integer vocabularies

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.text.TextPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.text.ConvertToVocabulary(
            {
                'test' : 3,
                'vocabulary' : 4
            }
        )
    )
    """

    def __init__(
        self,
        vocabulary: dict,
        start_character: int = 1,
        oov_character: int = 2,
        max_vocab: int = None
    ):
        """
        Parameters
        ----------
        vocabulary : dict
            Dictionary of string -> integer mappings
        start_character : int (default 1)
            The character to use for the start of an input sequence
        oov_character : int (default 2)
            The character to use for out of vocabulary tokens
        max_vocab : int or None (default None)
            The maximum vocabulary integer to use. If None, all vocabulary are used
        """
        super().__init__()
        self.vocabulary = vocabulary
        self.start_character = start_character
        self.oov_character = oov_character
        self.max_vocab = max_vocab

    @property
    def vocabulary(self):
        return self._vocabulary

    @vocabulary.setter
    def vocabulary(self, value):
        if not isinstance(value, dict):
            raise TypeError('vocabulary must be dictionary')
        if not all([isinstance(k, str) for k in value.keys()]):
            raise ValueError('All keys in vocabulary must be strings')
        if not all([isinstance(v, int) for v in value.values()]):
            raise ValueError('All values in vocabulary must be integers')
        self._vocabulary = value

    @property
    def start_character(self):
        return self._start_character

    @start_character.setter
    def start_character(self, value):
        if not isinstance(value, int):
            raise TypeError('start_character must be int')
        self._start_character = value

    @property
    def oov_character(self):
        return self._oov_character

    @oov_character.setter
    def oov_character(self, value):
        if not isinstance(value, int):
            raise TypeError('oov_character must be int')
        self._oov_character = value

    @property
    def max_vocab(self):
        return self._max_vocab

    @max_vocab.setter
    def max_vocab(self, value):
        if value is not None:
            if not isinstance(value, int):
                raise TypeError('max_vocab must be int')
        self._max_vocab = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ConvertToVocabulary',
            'params': {
                'vocabulary': self.vocabulary,
                'startCharacter': self.start_character,
                'oovCharacter': self.oov_character,
                'maxVocab': self.max_vocab
            }
        }


class PadSequences(BaseObject):
    """Text preprocessing object to pad sequences

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.text.TextPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.text.PadSequences()
    )
    """

    def __init__(
        self,
        pad_character: int = 0,
        length: int = 128,
        pad_location: str = 'post',
        truncate_location: str = 'post'
    ):
        """
        Parameters
        ----------
        pad_character : int (default 0)
            The character to use for padding
        length : int (default 128)
            The length to pad sequences to
        pad_location : str (default 'post')
            One of either 'pre' or 'post', corresponding to how
            sequences are to be padded
        truncate_location : str (default 'post')
            One of either 'pre' or 'post', corresponding to how
            sequences are to be truncated
        """
        super().__init__()
        self.pad_character = pad_character
        self.length = length
        self.pad_location = pad_location
        self.truncate_location = truncate_location

    @property
    def pad_character(self):
        return self._pad_character

    @pad_character.setter
    def pad_character(self, value):
        if not isinstance(value, int):
            raise TypeError('pad_character must be int')
        self._pad_character = value

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, value):
        if not isinstance(value, int):
            raise TypeError('length must be int')
        if value <= 0:
            raise ValueError('length must be greater than 0')
        self._length = value

    @property
    def pad_location(self):
        return self._pad_location

    @pad_location.setter
    def pad_location(self, value):
        if not isinstance(value, str):
            raise TypeError('pad_location must be str')
        if value not in ALLOWED_PADS:
            raise ValueError(f'pad_location must be one of {ALLOWED_PADS}')
        self._pad_location = value

    @property
    def truncate_location(self):
        return self._truncate_location

    @truncate_location.setter
    def truncate_location(self, value):
        if not isinstance(value, str):
            raise TypeError('truncate_location must be str')
        if value not in ALLOWED_PADS:
            raise ValueError(
                f'truncate_location must be one of {ALLOWED_PADS}')
        self._truncate_location = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'PadSequences',
            'params': {
                'padCharacter': self.pad_character,
                'length': self.length,
                'padLocation': self.pad_location,
                'truncateLocation': self.truncate_location
            }
        }


class Trim(BaseObject):
    """Text preprocessing class to trim whitespace from text

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.text.TextPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.text.Trim()
    )
    """

    def __init__(self):
        super().__init__()

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'Trim',
            'params': {}
        }



from typing import Type
import pytest
import aisquared


def test_postprocesser_init():
    with pytest.raises(TypeError):
        aisquared.config.postprocessing.BinaryClassification(
            ['zero', 'one'],
            0
        )
        aisquared.config.postprocessing.ObjectDetection(
            ['zero', 'one'],
            0
        )
        aisquared.config.postprocessing.Regression(
            min='test'
        )
        aisquared.config.postprocessing.Regression(
            max='test'
        )
        aisquared.config.postprocessing.Regression(
            round='test'
        )

    with pytest.raises(ValueError):
        aisquared.config.postprocessing.BinaryClassification(
            ['zero', 'one'],
            -0.5
        )
        aisquared.config.postprocessing.ObjectDetection(
            ['zero', 'one'],
            -0.5
        )
        aisquared.config.postprocessing.MulticlassClassification(
            ['zero', 'one']
        )

    aisquared.config.postprocessing.BinaryClassification(
        ['zero', 'one']
    )
    aisquared.config.postprocessing.ObjectDetection(
        ['zero', 'one']
    )
    aisquared.config.postprocessing.MulticlassClassification(
        ['zero', 'one', 'two']
    )
    aisquared.config.postprocessing.Regression(
        12,
        24,
        round=True
    )



from typing import Union
from aisquared.base import BaseObject


class AddValue(BaseObject):
    """
    Preprocessing step to add a value to all pixels in an image

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.image.ImagePreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.image.AddValue(255.0)
    )
    """

    def __init__(
            self,
            value: Union[int, float]
    ):
        """
        Parameters
        ----------
        value : int or float
            The value to add
        """
        super().__init__()
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError('value must be int or float')
        self._value = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'Add',
            'params': {
                'value': self.value
            }
        }


class SubtractValue(BaseObject):
    """
    Preprocessing step to subtract a value from all pixels in an image

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.image.ImagePreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.image.SubtractValue(255.0)
    )
    """

    def __init__(
            self,
            value: Union[int, float]
    ):
        """
        Parameters
        ----------
        value : int or float
            The value to subtract
        """
        super().__init__()
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError('value must be int or float')
        self._value = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'Subtract',
            'params': {
                'value': self.value
            }
        }


class MultiplyValue(BaseObject):
    """
    Preprocessing step to multiply all pixels in an image by a value

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.image.ImagePreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.image.MultiplyValue(2.0)
    )
    """

    def __init__(
            self,
            value: Union[int, float]
    ):
        """
        Parameters
        ----------
        value : int or float
            The value to multiply all pixels by
        """
        super().__init__()
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError('value must be int or float')
        self._value = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'Multiply',
            'params': {
                'value': self.value
            }
        }


class DivideValue(BaseObject):
    """
    Preprocessing step to divide all pixels in an image by a value

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.image.ImagePreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.image.DivideValue(255.0)
    )
    """

    def __init__(
            self,
            value: Union[int, float]
    ):
        """
        Parameters
        ----------
        value : int or float
            The value to divide all pixels by
        """
        super().__init__()
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError('value must be int or float')
        self._value = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'Divide',
            'params': {
                'value': self.value
            }
        }


class ConvertToColor(BaseObject):
    """
    Preprocessing step to convert images to a color scheme

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.image.ImagePreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.image.ConvertToColor('RGB')
    )
    """

    def __init__(self, color: str):
        """
        Parameters
        ----------
        color : str
            Either 'RGB' or 'B+W', each corresponding to RGB or grayscale color schemes, respectively
        """
        super().__init__()
        self.color = color

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        if not isinstance(value, str):
            raise TypeError('color must be string')
        if value not in ['RGB', 'B+W']:
            raise ValueError('color must be one of `RGB` or `B+W`')
        self._color = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ConvertToColor',
            'params': {
                'color': self.color
            }
        }


class Resize(BaseObject):
    """
    Preprocessing step to resize an image

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.image.ImagePreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.image.Resize([100, 100])
    )
    """

    def __init__(
            self,
            size: list,
            method: str = 'bilinear',
            preserve_aspect_ratio: bool = False
    ):
        """
        Parameters
        ----------
        size : list
            List of two integer values, supplied as [`height`, `width`], to resize images to
        method : str (default 'bilinear')
            The method to use for resizing. Must be one of 'bilinear', 'lanczos3', 'lanczos5', 'bicubic', 'gaussian',
            'nearest', or 'mitchellcubic'
        preserve_aspect_ratio : bool (default False)
            Whether to preserve aspect ratio when resizing
        """
        super().__init__()
        self.size = size
        self.method = method
        self.preserve_aspect_ratio = preserve_aspect_ratio

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value):
        if not isinstance(value, list):
            raise TypeError('size must be a list')
        if len(value) != 2 or not all([isinstance(val, int) for val in value]):
            raise ValueError('size must be a list of two integers')
        self._size = value

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, value):
        allowed_methods = [
            'bilinear',
            'lanczos3',
            'lanczos5',
            'bicubic',
            'gaussian',
            'nearest',
            'area',
            'mitchellcubic'
        ]
        if value not in allowed_methods:
            raise ValueError(f'method must be one of {allowed_methods}')
        self._method = value

    @property
    def preserve_aspect_ratio(self):
        return self._preserve_aspect_ratio

    @preserve_aspect_ratio.setter
    def preserve_aspect_ratio(self, value):
        if not isinstance(value, bool):
            raise TypeError('preserve_aspect_ratio must be bool')
        self._preserve_aspect_ratio = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'Resize',
            'params': {
                'size': self.size,
                'method': self.method,
                'preserveAspectRatio': self.preserve_aspect_ratio
            }
        }



from aisquared.base import BaseObject


class BinaryClassification(BaseObject):
    """
    Postprocesssing configuration object for binary classification

    Example usage

    >>> import aisquared
    >>> my_obj = aisquared.config.postprocessing.BinaryClassification(
        ['class1', 'class2']
    )
    >>> my_obj.to_dict()
    {'className': 'BinaryClassification',
    'params': {'labelMap': ['class1', 'class2'], 'threshold': 0.5}}

    """

    def __init__(
        self,
        label_map: list,
        threshold: float = 0.5
    ):
        """
        Parameters
        ----------
        label_map : list
            List of two values to be mapped to the model outputs
        threshold : float (default 0.5)
            The threshold for the second value to the label map to be the one chosen
        """
        super().__init__()
        self.label_map = label_map
        self.threshold = threshold

    @property
    def label_map(self):
        return self._label_map

    @label_map.setter
    def label_map(self, value):
        if not isinstance(value, list):
            raise TypeError('label_map must be a list')
        if len(value) != 2:
            raise ValueError(
                'label_map must have exactly two values for binary classification')
        self._label_map = value

    @property
    def threshold(self):
        return self._threshold

    @threshold.setter
    def threshold(self, value):
        if not isinstance(value, float):
            raise TypeError('threshold must be float-valued')
        if value < 0 or value > 1:
            raise ValueError('threshold value must be between 0 and 1')
        self._threshold = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'BinaryClassification',
            'params': {
                'labelMap': self.label_map,
                'threshold': self.threshold
            }
        }



from aisquared.base import BaseObject, DEFAULT_TABLE_RENDERING_CSS, TABLE_RENDERING_CSS_FILE
import platform
import json
import os


class TableRendering(BaseObject):
    """
    Class for rendering tables

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.TableRendering(
        'my table',
        'MyTableID',
        'MyContainerID',
        'name_key',
        'value_key',
        'name_values'
    )
    >>> my_obj.to_dict()
    {'className': 'TableRendering',
    'label': 'my table',
    'params': {'id': 'MyTableID',
    'containerId': 'MyContainerID',
    'predictionNameKey': 'name_key',
    'predictionValueKey': 'value_key',
    'predictionNameValues': 'name_values',
    'tableName': ''}}

    """

    def __init__(
        self,
        label: str,
        id: str,
        container_id: str,
        prediction_name_key: str,
        prediction_value_key: str,
        prediction_name_values: str,
        table_name: str = '',
        css_params: dict = None
    ):
        """
        Parameters
        ----------
        label : str
            Label for the table
        id : str
            ID for the table
        container_id : str
            The ID for the container
        prediction_name_key : str
            The key to use for the prediction name
        prediction_value_key : str
            The key to use for the prediction value
        prediction_name_values : list of str
            The name of the values for the prediction
        css_params : dict or None (default None)
            Additional CSS parameters
        """

        super().__init__()
        self.label = label
        self.id = id
        self.container_id = container_id
        self.prediction_name_key = prediction_name_key
        self.prediction_value_key = prediction_value_key
        self.prediction_name_values = prediction_name_values
        self.table_name = table_name

        if css_params is None:
            if os.path.exists(TABLE_RENDERING_CSS_FILE):
                with open(TABLE_RENDERING_CSS_FILE, 'r') as f:
                    self.css_params = json.load(f)
            else:
                self.css_params = DEFAULT_TABLE_RENDERING_CSS
        else:
            if css_params.get('style'):
                self.css_params = css_params
            else:
                self.css_params = {'style': css_params}

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'TableRendering',
            'label': self.label,
            'params': {
                'id': self.id,
                'containerId': self.container_id,
                'predictionNameKey': self.prediction_name_key,
                'predictionValueKey': self.prediction_value_key,
                'predictionNameValues': self.prediction_name_values,
                'tableName': self.table_name,
                'style': self.css_params['style']
            }
        }



"""
The aisquared.serving package contains utilities to serve models to a local REST endpoint.

Here is an example of how to serve a simple keras model using these utilities:

>>> # Assume model is already trained and stored in memory as model
>>> from aisquared import serving
>>> serving.save_keras_model(model, 'my_model')
>>> serving.deploy_model(
    'my_model',
    'keras',
    additional_functions_file = '<optional file containing `preprocess` and `postprocess` functions, if applicable>'
)
App created successfullly. Serving and awaiting requests

And to retrieve predictions from the model:

>>> # From a separate terminal, assume data is already loaded
>>> from aisquared import serving
>>> serving.get_remote_predictions(data) # Do not need to change host or port if predicting from the same machine
*predictions*
"""

try:
    from mlflow.tensorflow import save_model as save_tensorflow_model
    from mlflow.sklearn import save_model as save_sklearn_model
    from mlflow.pytorch import save_model as save_pytorch_model
    from mlflow.keras import save_model as save_keras_model
except ImportError:
    pass

from .deploy_model import deploy_model, load_beyondml_model, load_keras_model, load_pytorch_model, load_sklearn_model, load_tensorflow_model
from .get_remote_prediction import get_remote_prediction



from aisquared.base import BaseObject


class ImageHarvester(BaseObject):
    """
    Object to harvest images

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.harvesting.ImageHarvester()
    >>> my_obj.to_dict()
    {'className': 'ImageHarvester', 'params': {'how' : 'all'}}

    """

    def __init__(
        self,
        how: str = 'all'
    ):
        """
        Parameters
        ----------
        how : str (default 'all')
            Which images to harvest (CURRENTLY ONLY SUPPORTS 'all')
        """
        super().__init__()
        self.how = how

    @property
    def how(self):
        return self._how

    @how.setter
    def how(self, value):
        if value != 'all':
            raise ValueError(
                "Currently the only value supported for how is 'all'")
        self._how = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ImageHarvester',
            'params': {
                'how': self.how
            }
        }



import platform
import json
import os

DEFAULT_CONTAINER_RENDERING_CSS = {
    'style': {
        'container': {
            'color': 'black',
            'margin': '0',
            'padding': '24px',
            'fontSize': '16px',
            'background': 'white',
            'borderRadius': '10px'
        }
    }
}

DEFAULT_HTML_TAG_RENDERING_CSS = {
    'style': {
        'content': {
            'border': "none",
            'margin': "0",
            'padding': "8px",
            'background': "transparent",
            'borderRadius': "0",
            'color': "#000000",
            'fontSize': '14px',
            'textAlign': 'left',
            'textDecoration': 'none',
            'textTransform': 'none',
            'fontWeight': 'bold',
            'width': 'auto',
            'height': 'auto',
        },
        'extraContent': {
            'border': "none",
            'margin': "0",
            'padding': "8px",
            'background': "transparent",
            'borderRadius': "0",
            'color': "#000000",
            'fontSize': '14px',
            'textAlign': 'left',
            'textDecoration': 'none',
            'textTransform': 'none',
            'fontWeight': 'bold',
            'width': 'auto',
            'height': 'auto',
        },
    }
}

DEFAULT_TABLE_RENDERING_CSS = {
    'style': {
        'container': {
            'border': "1px solid lightgray",
            'margin': "8px 0",
            'padding': "10px",
            'background': "#f3f3f3",
            'borderRadius': "10px",
            'color': "#000000",
            'fontSize': '14px',
            'textAlign': 'left',
            'fontFamily': 'Raleway, sans-serif',
        },
        'title': {
            'fontWeight': 'bold',
            'marginBottom': '4px',
        },
        'table': {
            'border': "1px solid lightgray",
            'textAlign': 'center',
            'textTransform': 'capitalize',
            'width': "100%",
            'fontFamily': 'Raleway, sans-serif',
        },
        'tableHeader': {
            'border': "1px solid lightgray",
            'padding': "4px",
            'fontWeight': 'bold',
            'textTransform': 'capitalize',
        },
        'tableBody': {
        },
        'tableRows': {
            'border': "1px solid lightgray",
        },
        'rowCells': {
            'border': "1px solid lightgray",
            'fontWeight': 'normal',
            'padding': "8px",
        },
        'headerCells': {
            'border': '1px solid lightgray',
            'padding': '8px',
            'fontWeight': 'bolder',
            'backgroundColor': 'white'
        }
    }
}
DEFAULT_CHART_RENDERING_CSS = {
    "style": {
        "container": {
            "color": "black",
            "border": "1px solid lightgray",
            "margin": "0",
            "padding": "24px",
            "fontSize": "16px",
            "background": "white",
            "borderRadius": "10px"
        }
    }
}

if platform.system() == 'Windows':
    basedir = os.getenv('HOMEPATH')
else:
    basedir = os.getenv('HOME')

DIRECTORY = os.path.join(basedir, '.aisquared')

TABLE_RENDERING_CSS_FILE = os.path.join(DIRECTORY, 'TableRendering.json')
HTML_TAG_RENDERING_CSS_FILE = os.path.join(DIRECTORY, 'HTMLTagRendering.json')
CHART_RENDERING_CSS_FILE = os.path.join(DIRECTORY, 'ChartRendering.json')
CONTAINER_RENDERING_CSS_FILE = os.path.join(
    DIRECTORY, 'ContainerRendering.json')



from aisquared.base import BaseObject


class MulticlassFeedback(BaseObject):
    """
    Feedback for multiclass classification

    Example Usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.feedback.MulticlassFeedback(['class1', 'class2', 'class3'])
    >>> my_obj.to_dict()
    {'className': 'MulticlassFeedback',
    'params': {'labelMap': ['class1', 'class2', 'class3']}}
    """

    def __init__(self, label_map: list):
        """
        Parameters
        ----------
        label_map : list
            List of class values
        """
        super().__init__()
        self.label_map = label_map

    @property
    def label_map(self):
        return self._label_map

    @label_map.setter
    def label_map(self, value):
        if not isinstance(value, list):
            raise TypeError('label_map must be list')
        self._label_map = value

    def to_dict(self) -> dict:
        """
        Return the object as a dictionary
        """
        return {
            'className': 'MulticlassFeedback',
            'params': {
                'labelMap': self.label_map
            }
        }



from aisquared.base import BaseObject


class TextRendering(BaseObject):
    """
    Class for rendering text

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.TextRendering(
        prediction_value_key = 'my_key'
    )
    >>> my_obj.to_dict()
    {'className': 'TextRendering',
    'params': {'predictionValueKey': 'my_key'}}
    """

    def __init__(
            self,
            prediction_value_key: str = None
    ):
        """
        Parameters
        ----------
        prediction_value_key : str or None (default None)
            If present, the key in the prediction to render
        """
        super().__init__()
        self.prediction_value_key = prediction_value_key

    @property
    def prediction_value_key(self):
        return self._prediction_value_key

    @prediction_value_key.setter
    def prediction_value_key(self, value):
        if value is not None:
            if not isinstance(value, str):
                raise TypeError(
                    f'prediction_value_key must be str, got {type(value)}')
        self._prediction_value_key = value

    def to_dict(self):
        return {
            'className': 'TextRendering',
            'params': {
                'predictionValueKey': self.prediction_value_key
            }
        }



"""
The aisquared.config subpackage contains utilities and objects for packaging aisquared configuration steps and models.

For in-depth examples of how to build out .air files using the utilities and classes in this library, please visit our
GitHub repository at https://github.com/AISquaredInc/airFiles
"""

from .ModelConfiguration import ModelConfiguration
from .GraphConfiguration import GraphConfiguration
from .CustomObject import CustomObject



from .AISquaredAPIException import AISquaredAPIException
from .additional_utils import _check_results_length
from aisquared.base import ENDPOINTS
import pandas as pd
import requests
import json


def _list_models(
        url: str,
        headers: dict,
        as_df: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Format a URL based on parameters for whether the user is interacting with the ALB or not

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    as_df : bool
        Whether to return the response as a dataframe object
    """

    url = f'{url}/{ENDPOINTS["model"]}?userOnly=true'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )
    if not resp.ok:
        raise AISquaredAPIException(json.dumps(resp.json()))

    else:
        if as_df:
            df = pd.DataFrame(resp.json()['data']['models'])
            _check_results_length(df)
            df['name'] = df.config.apply(lambda c: c['params']['name'])
            new_cols = ['name', 'id']
            new_cols += [c for c in df.columns if c not in new_cols]

            return df[new_cols]

        return resp.json()['data']['models']


def _upload_model(
        url: str,
        headers: dict,
        model_file: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Upload a model to the platform

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    model_file : path or path-like
        Whether to return the response as a dataframe object
    """

    url = f'{url}/{ENDPOINTS["upload_model"]}'

    with open(model_file, 'rb') as f:

        with requests.Session() as sess:
            resp = sess.post(
                url,
                headers=headers,
                files={'model': f}
            )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    return resp.json()['data']['id']


def _get_model(
        url: str,
        headers: dict,
        model_id: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Upload a model to the platform

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    model_id : string
        The id of the model you want to interact with
    """

    url = f'{url}/{ENDPOINTS["model"]}/{model_id}'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    return resp.json()['data']


def _delete_model(
        url: str,
        headers: dict,
        model_id: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Upload a model to the platform

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    model_id : string
        The id of the model you want to interact with
    """

    url = f'{url}/{ENDPOINTS["model"]}/{model_id}'

    with requests.Session() as sess:
        resp = sess.delete(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    return resp.ok



from aisquared.base import LOCATIONS, COLORS, BaseObject


class ObjectRendering(BaseObject):
    """
    Object which dictates how to render object detection in images

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.ObjectRendering()
    >>> my_obj.to_dict()
    {'className': 'ObjectRendering',
    'params': {'color': 'blue',
    'thickness': '5',
    'placement': 'bottomleft',
    'includeProbability': False,
    'badgeColor': 'white',
    'fontColor': 'black',
    'fontSize': '5'}}

    """

    def __init__(
        self,
        color: str = COLORS[-1],
        thickness: str = '5',
        placement: str = LOCATIONS[-1],
        include_probability: bool = False,
        badge_color: str = COLORS[-2],
        font_color: str = COLORS[-4],
        font_size: str = '5'
    ):
        """
        Parameters
        ----------
        color : str (default 'blue')
            The color of the box around images
        thickness : str (default '5')
            The thickness of the box around images
        placement : str (default 'bottomleft')
            The placement of the prediction text
        include_probability : bool (default False)
            Whether to render calculated probabilities
        badge_color : str (default 'white')
            Background color for the text region
        font_color : str (default 'black')
            Color of the text
        font_size : str (default '5')
            Text size
        """
        super().__init__()
        self.color = color
        self.thickness = thickness
        self.placement = placement
        self.include_probability = include_probability
        self.badge_color = badge_color
        self.font_color = font_color
        self.font_size = font_size

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        if value not in COLORS:
            raise ValueError(f'color must be one of {COLORS}')
        self._color = value

    @property
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, value):
        self._thickness = value

    @property
    def placement(self):
        return self._placement

    @placement.setter
    def placement(self, value):
        if value not in LOCATIONS:
            raise ValueError(f'placement must be one of {LOCATIONS}')
        self._placement = value

    @property
    def include_probability(self):
        return self._include_probability

    @include_probability.setter
    def include_probability(self, value):
        self._include_probability = value

    @property
    def badge_color(self):
        return self._badge_color

    @badge_color.setter
    def badge_color(self, value):
        if value not in COLORS:
            raise ValueError(f'badge_color must be one of {COLORS}')
        self._badge_color = value

    @property
    def font_color(self):
        return self._font_color

    @font_color.setter
    def font_color(self, value):
        if value not in COLORS:
            raise ValueError(f'font_color must be one of {COLORS}')
        self._font_color = value

    @property
    def font_size(self):
        return self._font_size

    @font_size.setter
    def font_size(self, value):
        self._font_size = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ObjectRendering',
            'params': {
                'color': self.color,
                'thickness': self.thickness,
                'placement': self.placement,
                'includeProbability': self.include_probability,
                'badgeColor': self.badge_color,
                'fontColor': self.font_color,
                'fontSize': self.font_size
            }
        }



from aisquared.base import BaseObject


class SimpleFeedback(BaseObject):
    """
    Simple thumbs-up/thumbs-down feedback for predictions

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.feedback.SimpleFeedback()
    >>> my_obj.to_dict()
    {'className': 'SimpleFeedback', 'params': {}}
    """

    def __init__(self):
        super().__init__()

    def to_dict(self) -> dict:
        """
        Return the object as a dictionary
        """
        return {
            'className': 'SimpleFeedback',
            'params': dict()
        }



from setuptools import setup
from aisquared import __version__

with open('requirements.txt', 'r') as f:
    requirements = [line for line in f.read().splitlines() if line != '']
with open('additional_requirements.txt', 'r') as f:
    additional_requirements = [
        line for line in f.read().splitlines() if line != '']

setup(
    name='aisquared',
    version=__version__,
    url='https://github.com/AISquaredInc/aisquared',
    packages=[
        'aisquared',
        'aisquared.base',
        'aisquared.config',
        'aisquared.logging',
        'aisquared.serving',
        'aisquared.config.harvesting',
        'aisquared.config.preprocessing',
        'aisquared.config.preprocessing.text',
        'aisquared.config.preprocessing.image',
        'aisquared.config.preprocessing.tabular',
        'aisquared.config.analytic',
        'aisquared.config.postprocessing',
        'aisquared.config.rendering',
        'aisquared.config.feedback',
        'aisquared.utils',
        'aisquared.platform'
    ],
    author='The AI Squared Team',
    author_email='staff@squared.ai',
    description='Utilities for interacting with the AI Squared Technology Stack',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    license_files='LICENSE',
    install_requires=requirements,
    extras_require={'full': additional_requirements},
)



from aisquared.base import BaseObject


class MulticlassClassification(BaseObject):
    """
    Postprocessing configuration object for multiclass classification

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.postprocessing.MulticlassClassification(
        ['class1', 'class2', 'class3']
    )
    >>> my_obj.to_dict()
    {'className': 'MulticlassClassification',
    'params': {'labelMap': ['class1', 'class2', 'class3']}}

    """

    def __init__(
            self,
            label_map: list,
    ):
        """
        Parameters
        ----------
        label_map : list
            A list of values to map the output of the model to
        """
        super().__init__()
        self.label_map = label_map

    @property
    def label_map(self):
        return self._label_map

    @label_map.setter
    def label_map(self, value):
        if not isinstance(value, list):
            raise ValueError('label_map must be a list')
        if len(value) <= 2:
            raise ValueError(
                'For multiclass classification, the label map must have more than two values. If there are only two values, use the `BinaryClassification` class')
        self._label_map = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'MulticlassClassification',
            'params': {
                'labelMap': self.label_map
            }
        }



from aisquared.base import BaseObject


class ChatbotHarvester(BaseObject):
    """
    Harvesting for a chatbot
    """

    def __init__(
            self,
            title,
            harvest_history=False,
            input_type='text',
            features=None,
            max_length=None
    ):
        """
        Parameters
        ----------
        title : str
            The title for the chat conversation
        harvest_history : bool (default False)
            Whether to harvest chat history or just the last sent text
        input_type : str (default 'text')
            The input type (should not be changed)
        features : default None
            Should not be changed
        max_length : default None
            Should not be changed
        """
        super().__init__()
        self.title = title
        self.harvest_history = harvest_history
        self.input_type = input_type
        self.features = features
        self.max_length = max_length

    def to_dict(self):
        """
        Return the configuration object as a dictionary
        """
        return {
            'className': 'ChatbotHarvester',
            'params': {
                'title': self.title,
                'harvestHistory': self.harvest_history,
                'inputType': self.input_type,
                'features': self.features,
                'maxLength': self.max_length
            }
        }



import os
import json
import pytest
import aisquared
import tensorflow as tf


def test_graphconfig(tmp_path):
    regex = '\D(\d{5})\D'
    analytic = {
        '11111': {
            'name': 'John Doe',
            'nbo': 1
        },
        '22222': {
            'name': 'Jane Doe',
            'nbo': 1
        },
        '33333': {
            'name': 'Alice',
            'nbo': 3
        },
        '44444': {
            'name': 'Bob',
            'nbo': 8
        }
    }
    with open(os.path.join(tmp_path, 'analytic.json'), 'w') as f:
        json.dump(analytic, f)

    harvester = aisquared.config.harvesting.TextHarvester('regex', regex)
    preproc = None
    analytic = aisquared.config.analytic.LocalAnalytic(
        os.path.join(tmp_path, 'analytic.json'), input_type='text')
    postproc = None
    rendering = aisquared.config.rendering.WordRendering()
    model_feedback = aisquared.config.feedback.ModelFeedback()
    model_feedback.add_question(
        'Is this model helpful?', 'singleChoice', ['yes', 'no'])
    prediction_feedback = aisquared.config.feedback.SimpleFeedback()
    feedback = prediction_feedback

    config = aisquared.config.GraphConfiguration('TextAnalytic')
    harvester_id = config.add_node(harvester)
    analytic_id = config.add_node(analytic, harvester_id)
    render_id = config.add_node(rendering, analytic_id)
    model_feedback_id = config.add_node(model_feedback, render_id)
    prediction_feedback_id = config.add_node(prediction_feedback, render_id)

    config.compile(os.path.join(tmp_path, 'TextAnalytic.air'))
    assert True


def test_sentimentanalysis(tmp_path):
    input_layer = tf.keras.layers.Input(128)
    x = tf.keras.layers.Embedding(100, 2)(input_layer)
    x = tf.keras.layers.Flatten()(x)
    x = tf.keras.layers.Dense(100, activation='relu')(x)
    output_layer = tf.keras.layers.Dense(1, activation='relu')(x)

    model = tf.keras.models.Model(input_layer, output_layer)
    model.save(os.path.join(tmp_path, 'model.keras'))

    harvester = aisquared.config.harvesting.TextHarvester()
    preproc = aisquared.config.preprocessing.text.TextPreprocesser(
        [
            aisquared.config.preprocessing.text.Tokenize(),
            aisquared.config.preprocessing.text.ConvertToVocabulary(
                {
                    'this': 3,
                    'is': 4,
                    'a': 5,
                    'test': 6
                }
            ),
            aisquared.config.preprocessing.text.PadSequences()
        ]
    )
    analytic = aisquared.config.analytic.LocalModel(
        os.path.join(tmp_path, 'model.keras'), input_type='text')
    postproc = aisquared.config.postprocessing.Regression()
    rendering = aisquared.config.rendering.DocumentRendering()
    feedback = aisquared.config.feedback.RegressionFeedback()

    config = aisquared.config.ModelConfiguration(
        'sentiment-analysis',
        harvester,
        preproc,
        analytic,
        postproc,
        rendering,
        feedback_steps=feedback,
        url='https://dynamics.microsoft.com/'
    )
    config.compile(os.path.join(tmp_path, 'sentiment-analysis.air'))
    assert True


def test_imageclassification(tmp_path):
    input_layer = tf.keras.layers.Input((30, 30, 3))
    x = tf.keras.layers.Flatten()(input_layer)
    x = tf.keras.layers.Dense(1, activation='sigmoid')(x)
    model = tf.keras.models.Model(input_layer, x)
    model.save(os.path.join(tmp_path, 'test_model.keras'))

    harvester = aisquared.config.harvesting.InputHarvester('image')
    preprocesser = aisquared.config.preprocessing.image.ImagePreprocesser(
        [
            aisquared.config.preprocessing.image.Resize([30, 30]),
            aisquared.config.preprocessing.image.DivideValue(255)
        ]
    )
    analytic = aisquared.config.analytic.LocalModel(
        os.path.join(tmp_path, 'test_model.keras'), 'cv')
    postprocesser = aisquared.config.postprocessing.BinaryClassification(
        [
            'zero',
            'one'
        ]
    )
    renderer = aisquared.config.rendering.DocumentRendering()
    config = aisquared.config.ModelConfiguration(
        'InputHarvestTest',
        harvester,
        preprocesser,
        analytic,
        postprocesser,
        renderer
    )
    config.compile(os.path.join(tmp_path, config.name))



try:
    from mlflow.tensorflow import load_model as load_tensorflow_model
    from mlflow.sklearn import load_model as load_sklearn_model
    from mlflow.pytorch import load_model as load_pytorch_model
    from mlflow.keras import load_model as load_keras_model
except ImportError:
    pass

try:
    from flask import Flask, request, Response
except ImportError:
    pass

import tensorflow as tf
import numpy as np

try:
    import waitress
except ImportError:
    pass

try:
    import torch
except ImportError:
    pass

from importlib import import_module
import json
import os

_ALLOWED_TYPES = [
    'tensorflow',
    'sklearn',
    'pytorch',
    'keras',
    'beyondml'
]


def load_beyondml_model(model: str, custom_objects: dict):
    """
    Load a BeyondML model with custom objects
    """
    return tf.keras.models.load_model(
        model,
        custom_objects=custom_objects
    )


def deploy_model(
        saved_model: str,
        model_type: str,
        host: str = '127.0.0.1',
        port: int = 2244,
        custom_objects: dict = None,
        additional_functions_file: str = None
):
    """
    Deploy a model to a Flask server on the specified host

    Parameters
    ----------
    saved_model : Path-like
        The path to the saved model directory or model file
    model_type : str
        The type of model
    host : str (default '127.0.0.1')
        The host to deploy to
    port : int (default 2244)
        The port to deploy to
    custom_objects : dict or None (default None)
        Any custom objects to load when using a BeyondML model
    additional_functions_file : file-like or None (default None)
        File name containing additional functions (which have to be named `preprocess` and `postprocess`, if created)
        that are used during the prediction process
    """
    if model_type not in _ALLOWED_TYPES:
        raise ValueError(
            f'model_type must be one of {_ALLOWED_TYPES}, got {model_type}')

    if model_type == 'tensorflow':
        model = load_tensorflow_model(saved_model)
    elif model_type == 'sklearn':
        model = load_sklearn_model(saved_model)
    elif model_type == 'pytorch':
        model = load_pytorch_model(saved_model)
    elif model_type == 'keras':
        model = load_keras_model(saved_model)
    elif model_type == 'beyondml':
        model = load_beyondml_model(saved_model, custom_objects)

    # Import preprocessing and postprocessing steps, if provided
    if additional_functions_file:
        file_name = os.path.splitext(
            os.path.basename(additional_functions_file))[0]
        dir_name = os.path.dirname(os.path.abspath(additional_functions_file))
        module = import_module(file_name, dir_name)

        try:
            preprocess = module.preprocess
        except AttributeError:
            preprocess = None

        try:
            postprocess = module.postprocess
        except AttributeError:
            postprocess = None

    # Create the Flask app
    app = Flask(__name__)

    # Create the predict function
    @app.route('/predict', methods=['POST'])
    def predict():

        # try to get the data
        try:
            data = request.get_json()
            to_predict = data['data']
            if preprocess:
                to_predict = preprocess(to_predict)
        except Exception:
            return Response(
                'Data appears to be incorrectly formatted',
                400
            )

        # try to get the data correctly formatted for prediction
        try:
            if model_type == 'beyondml':
                to_predict = [
                    np.asarray(d) for d in to_predict
                ]
            elif model_type in ['tensorflow', 'keras', 'pytorch', 'sklearn']:
                to_predict = np.asarray(to_predict)
        except Exception:
            return Response(
                'Data passed could not be correctly converted to numpy array for prediction',
                400
            )

        # try to return the actual predictions
        try:
            if model_type != 'pytorch':
                predictions = np.asarray(model.predict(to_predict)).tolist()
            elif model_type == 'pytorch':
                predictions = model(torch.Tensor(to_predict)
                                    ).detach().numpy().tolist()

            if postprocess:
                try:
                    predictions = postprocess(predictions)
                except Exception as e:
                    print(e)
            return json.dumps({
                'predictions': predictions
            })

        except Exception:
            return Response(
                'Error in performing prediction',
                400
            )

    # run the app
    print('App created successfully. Serving and awaiting requests.')
    waitress.serve(
        app,
        host=host,
        port=port
    )



"""
This package contains utilities to interact with the AI Squared technology stack, particularly with developing and deploying models to the AI Squared Browser Extension or other applications developed through the AI Squared JavaScript SDK.
"""

__version__ = '0.3.12'
__dev__ = True

import aisquared.config
import aisquared.base
import aisquared.platform

try:
    import aisquared.logging
    import aisquared.serving
    import aisquared.utils
    import llmlink
except ImportError:
    pass



from typing import Union

from getpass import getpass
import pandas as pd
import warnings
import requests
import json
import os

from aisquared.base import DIRECTORY, CLIENT_CONFIG_FILE, ENDPOINTS

from .AISquaredAPIException import AISquaredAPIException
from .crudl import _list_models, _upload_model, _get_model, _delete_model
from .sharing import _list_model_users, _model_share_with_user, _model_share_with_group
from .feedback import _list_model_feedback, _list_prediction_feedback, _list_model_prediction_feedback
from .user_group import _create_user, _update_user, _delete_user, _get_user, _get_group, _create_group, _delete_group, _update_group, _users_to_group, _list_users, _list_groups, _list_group_users, _list_roles
from .metrics import _list_user_usage_metrics, _list_model_usage_metrics
from .additional_utils import _test_connection


class AISquaredPlatformClient:
    """
    Client for interacting with the AI Squared platform programmatically

    When using the client for the first time, it is important to run the `client.login()` method. When doing so, the
    client will ask for any required information interactively.

    >>> import aisquared
    >>> client = aisquared.platform.AISquaredPlatformClient()
    >>> # If you have never logged in before, run the following code:
    >>> client.login()
    >>> # Test connection
    >>> client.test_connection()
    True

    """

    def __init__(
        self,
        use_port: bool = False
    ):
        """
        Parameters
        ----------
        use_port : bool (default False)
            Whether to default to using the port parameter for all API calls
        """

        try:
            self._load_info(CLIENT_CONFIG_FILE)
        except Exception as e:
            warnings.warn(
                'It appears you are not authenticated to the AI Squared Platform. Please run Client.login() before performing any action'
            )

        self.use_port = use_port

    @property
    def use_port(self):
        return self._use_port

    @use_port.setter
    def use_port(self, value):
        if not isinstance(value, bool):
            raise TypeError('use_port must be Boolean')
        self._use_port = value

    def login(
        self,
        url: str = None,
        port: int = 8080,
        username: str = None,
        password: str = None,
        use_port: bool = None
    ) -> None:
        """
        Log in to the platform programmatically.  If no url, username, or password are provided, logs in interactively

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.login()
        Enter URL: https://platform.squared.ai
        Enter Username: your.email@your_domain.com
        Enter Password: <hidden>

        Parameters
        ----------
        url : str or None (default None)
            The URL for the platform API
        port : int or None (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        username : str or None (default None)
            The username
        password : str or None (default None)
            The password
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        """
        if url is None:
            url = input('Enter URL: ')
        if username is None:
            username = input('Enter Username: ')
        if password is None:
            password = getpass('Enter Password: ')

        if use_port is None:
            use_port = self.use_port

        call_url = self._format_url(url, port, use_port)

        with requests.Session() as sess:
            resp = sess.post(
                f'{call_url}/{ENDPOINTS["login"]}',
                data={
                    'username': username,
                    'password': password
                }
            )

        if not resp.ok:
            raise AISquaredAPIException('Authentication failed')
        else:
            token = resp.json()['token']['access_token']

        if not os.path.exists(DIRECTORY):
            os.makedirs(DIRECTORY)

        with open(CLIENT_CONFIG_FILE, 'w') as f:
            json.dump(
                {
                    'url': url,
                    'username': username,
                    'password': password,
                    'token': token
                },
                f
            )

        self._load_info()

    def _load_info(self, config_file: str = CLIENT_CONFIG_FILE) -> None:
        """
        NOT MEANT TO BE CALLED BY THE USER

        Function to load configuration information for the client
        """
        with open(config_file, 'r') as f:
            data = json.load(f)
        self._base_url = data['url']
        self._username = data['username']
        self._password = data['password']
        self._token = data['token']

    def _format_url(self, url: str, port: int, use_port: bool):
        """
        NOT MEANT TO BE CALLED BY THE END USER

        Format a URL based on parameters for whether the user is interacting with the ALB or not

        Parameters
        ----------
        url : string
            The base url to format
        port : int
            The port to use
        use_port : bool
            Whether to use the port

        Returns
        -------
        formatted_url : str
            The formatted URL
        """

        if use_port:
            return f'{url}:{port}'
        else:
            return url

    @property
    def headers(self):
        """Headers used for authentication with the AI Squared Platform"""
        return {
            'authorization': f'Bearer {self._token}',
            'authType': 'jwt',
            'User-Agent': 'AI Squared/1.0'
        }

    @property
    def username(self) -> str:
        """The username associated with the client"""
        return self._username

    @property
    def password(self) -> str:
        """The password associated with the client"""
        return '*' * len(self._password)

    @property
    def token(self) -> str:
        """The token associated with the client"""
        return '*' * len(self._token)

    @property
    def base_url(self) -> str:
        """The base URL associated with the client"""
        return self._base_url

    # CRUDL operations for models

    def list_models(self, as_df: bool = True, port: int = 8080, use_port: bool = None) -> Union[pd.DataFrame, dict]:
        """
        List models within the platform

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.list_models()
        *DataFrame with results*

        Parameters
        ----------
        as_df : bool (default True)
            Whether to return the response as a pandas DataFrame
        port : default None
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        models : pandas DataFrame or dictionary
            The models

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_models(url, self.headers, as_df)

    def upload_model(self, model_file: str, port: int = 8081, use_port: bool = None) -> str:
        """
        Upload a model to the platform

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.upload_model('my_model_filename.air')
        True

        Parameters
        ----------
        model_file : path or path-like
            The path to the model file
        port : int (default 8081)
            The API port to use. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        successful : bool
            Whether the action was successful

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _upload_model(
            url,
            self.headers,
            model_file
        )

    def get_model(self, id: str, port: int = 8080, use_port: bool = None) -> dict:
        """
        Retrieve a model configuration

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.get_model('model_id')
        *JSON Response including model data and metadata*

        Parameters
        ----------
        id : str
            The ID for the model
        port : int (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        model : dictionary
            Metadata about the model coupled with the model's configuration information

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _get_model(url, self.headers, id)

    def delete_model(self, id: str, port: int = 8080, use_port: bool = None) -> bool:
        """
        Delete a model

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.delete_model('model_id')
        True

        Parameters
        ----------
        id : str
            The ID for the model
        port : int (default 8080)
            The API port for the model. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        success : bool
            Whether the action was successful

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _delete_model(url, self.headers, id)

    # Sharing operations for models

    def list_model_users(self, id: str, as_df: bool = True, port: int = 8080, use_port: bool = None) -> Union[pd.DataFrame, dict]:
        """
        List users for a model

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.list_model_users('model_id')
        *DataFrame with results*

        Parameters
        ----------
        id : str
            The ID for the model
        as_df : bool (default True)
            Whether to return the response as a Pandas DataFrame
        port : int (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        model_users : pandas DataFrame or dictionary
            The users for the model

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_model_users(url, self.headers, id, as_df)

    def share_model_with_user(self, model_id: str, user_id: str, port: int = 8080, use_port: bool = None) -> bool:
        """
        Share a model with a user

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.share_model_with_user('model_id', 'user_id')
        True

        Parameters
        ----------
        model_id : str
            The ID for the model
        user_id : str
            The ID for the user
        port : int (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        success : bool
            Whether the action was successful

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _model_share_with_user(
            url,
            self.headers,
            model_id,
            user_id,
            True
        )

    def unshare_model_with_user(self, model_id: str, user_id: str, port: int = 8080, use_port: bool = None) -> bool:
        """
        Unshare a model with a user

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.unshare_model_with_user('model_id', 'user_id')
        True

        Parameters
        ----------
        model_id : str
            The ID for the model
        user_id : str
            The ID for the user
        port : int (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        success : bool
            Whether the action was successful

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _model_share_with_user(
            url,
            self.headers,
            model_id,
            user_id,
            False
        )

    def share_model_with_group(self, model_id: str, group_id: str, port: int = 8080, use_port: bool = None) -> bool:
        """
        Share a model with a group

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.share_model_with_group('model_id', 'group_id')
        True

        Parameters
        ----------
        model_id : str
            The ID for the model to be shared
        group_id : str
            The ID for the group to be shared with. This can be handled automatically by the platform ALB
        port : int (default 8080)
            The API port to use. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value


        Returns
        -------
        success : bool
            Returns True if successful
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _model_share_with_group(
            url,
            self.headers,
            model_id,
            group_id,
            True
        )

    def unshare_model_with_group(self, model_id: str, group_id: str, port: int = 8080, use_port: bool = None) -> bool:
        """
        Unshare a model with a group

        >>> import aisquared
        >>> client = aisquared.client.AISquaredPlatformClient()
        >>> client.unshare_model_with_group('model_id', 'group_id')
        True

        Parameters
        ----------
        model_id : str
            The ID of the model
        group_id : str
            The ID of the group
        port : int (default 8080)
            The API port to use. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        success : bool
            Returns True if successful
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _model_share_with_group(
            url,
            self.headers,
            model_id,
            group_id,
            False
        )

    # Feedback operations

    def list_model_feedback(self, model_id: str, limit: int = 10, as_df: bool = True, port: int = 8080, use_port: bool = None) -> Union[dict, pd.DataFrame]:
        """
        List feedback on a model

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.list_model_feedback('model_id')
        *DataFrame with Results*

        Parameters
        ----------
        model_id : str
            The ID of the model
        limit : int (default 10)
            The maximum number of feedback items to return
        port : int (default 8080)
            The API port to use. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        feedback : dict or pandas DataFrame
            The feedback
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_model_feedback(
            url,
            self.headers,
            model_id,
            limit,
            as_df
        )

    def list_prediction_feedback(self, prediction_id: str, as_df: bool = True, port: int = 8080, use_port: bool = None) -> Union[pd.DataFrame, dict]:
        """
        List prediction feedback given a prediction ID

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.list_prediction_feedback('prediction_id')
        *DataFrame with results*

        Parameters
        ----------
        prediction_id : str
            The prediction ID
        as_df : bool (default True)
            Whether to return the results as a pandas DataFrame
        port : int (default 8080)
            The API port to use. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        results : pandas DataFrame or dict
            The results from the platform

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_prediction_feedback(
            url,
            self.headers,
            prediction_id,
            as_df
        )

    def list_model_prediction_feedback(self, model_id: str, as_df: bool = True, port: int = 8080, use_port: bool = None) -> Union[dict, pd.DataFrame]:
        """
        List all feedback for a model

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.list_model_prediction_feedback('model_id')
        *DataFrame with Results*

        Parameters
        ----------
        model_id : str
            The ID of the model requested
        as_df : bool (default True)
            Whether to return the results as a pandas DataFrame
        port : int (default 8080)
            The API port to use. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        results : dict or pandas DataFrame
            The results from the platform
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_model_prediction_feedback(
            url,
            self.headers,
            model_id,
            as_df
        )

    # User and group management

    def create_user(
            self,
            user_name: str,
            given_name: str,
            family_name: str,
            email: str,
            role_id: str,
            active: bool = True,
            middle_name: str = None,
            company_id: str = None,
            password: str = None,
            port: int = 8085,
            use_port: bool = None
    ) -> dict:
        """
        Create a user within the platform

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.create_user(
            'user name',
            'given_name',
            'family_name',
            'user_email',
            'role_id'
        )
        *Dictionary with user information*

        Parameters
        ----------
        user_name : str
            The display name of the user
        given_name : str
            The user's first name
        family_name : str
            The user's last name
        email : str
            The user's email
        role_id : str
            The ID of the role to be given to the user
        active : bool (default True)
            Whether the user is active
        middle_name : str or None (default None)
            The user's middle name
        company_id : str or None (default None)
            The user's company ID
        password : str or None (default None)
            The user's password
        port : int (default 8085)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        user_data : dict
            Metadata about the user
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _create_user(
            url,
            self.headers,
            user_name,
            given_name,
            family_name,
            email,
            role_id,
            active,
            middle_name,
            company_id,
            password
        )

    def update_user(
            self,
            user_id: str,
            user_name: str,
            given_name: str,
            family_name: str,
            email: str,
            role_id: str,
            active: bool = True,
            middle_name: str = None,
            company_id: str = None,
            password: str = None,
            port: int = 8085,
            use_port: bool = None
    ) -> bool:
        """
        Update information about a user

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.update_user(
            'user_id',
            'user name',
            'given_name',
            'family_name',
            'user_email',
            'role_id'
        )
        True

        Parameters
        ----------
        user_id : str
            The ID of the user to update
        user_name : str
            The display name of the user
        given_name : str
            The first name of the user
        family_name : str
            The last name of the user
        email : str
            The user's email
        role_id : str
            The ID of the user's role
        active : bool (default True)
            Whether the user is active
        middle_name : str or None (default None)
            The user's middle name
        company_id : str or None (default None)
            The user's company ID
        password : str or None (default None)
            The user's password
        port : int (default 8085)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        success : bool
            Returns True if update is successful
        """

        json_data = {
            'active': active,
            'userName': user_name,
            'givenName': given_name,
            'familyName': family_name,
            'email': email,
            'roleId': role_id
        }

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _update_user(
            url,
            self.headers,
            user_id,
            user_name,
            given_name,
            family_name,
            email,
            role_id,
            active,
            middle_name,
            company_id,
            password
        )

    def delete_user(self, user_id: str, port: int = 8085, use_port: bool = None) -> bool:
        """
        Delete a user from the system

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.delete_user('user_id')
        True

        Parameters
        ----------
        user_id : str
            The user's ID
        port : int (default 8085)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        result : bool
            Returns True if the call is successful
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _delete_user(
            url,
            self.headers,
            user_id
        )

    def get_user(self, user_id: str, port: int = 8085, use_port: bool = None) -> dict:
        """
        Retrieve a user's information from the platform

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.get_user('user_id')
        *dictionary with results*

        Parameters
        ----------
        user_id: str
            The ID of the user
        port : int (default 8085)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        user_info : dict
            The information about the user
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _get_user(url, self.headers, user_id)

    def get_group(self, group_id: str, port: int = 8086, use_port: bool = None) -> dict:
        """
        Retrieve information about a group

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.get_group('group_id')
        *dictionary containing group data*

        Parameters
        ----------
        group_id : str
            The ID of the group requested
        port : int (default 8086)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        group_info : dict
            The information about the group
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _get_group(
            url,
            self.headers,
            group_id
        )

    def create_group(self, display_name: str, role_id: str, port: int = 8086, use_port: bool = None) -> dict:
        """
        Create a group in the platform

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.create_group(
            'group display name',
            'role_id'
        )
        *dictionary containing group information*

        Parameters
        ----------
        display_name : str
            The display name of the group
        role_id : str
            The role ID for the group
        port : int (default 8086)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value


        Returns
        -------
        group_info : dict
            Metadata about the created group
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _create_group(
            url,
            self.headers,
            display_name,
            role_id
        )

    def delete_group(self, group_id, port=8086, use_port: bool = None) -> bool:
        """
        Delete a group from the platform

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.delete_group('group_id')
        True

        Parameters
        ----------
        group_id : str
            The ID of the group to delete
        port : int (default 8086)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        result : bool
            Returns True if successful
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _delete_group(
            url,
            self.headers,
            group_id
        )

    def update_group(self, group_id: str, display_name: str, role_id: str, port: int = 8086, use_port: bool = None) -> bool:
        """
        Update information about a group

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.update_group(
            'group_id',
            'group display name',
            'role_id'
        )
        True

        Parameters
        ----------
        group_id : str
            The ID of the group to update
        display_name : str
            The display name of the group
        role_id : str
            The ID of the role for the group
        port : int (default 8086)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        success : bool
            Returns True if successful
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _update_group(
            url,
            self.headers,
            group_id,
            display_name,
            role_id
        )

    def add_users_to_group(self, group_id: str, user_ids: list, port: int = 8086, use_port: bool = None) -> bool:
        """
        Add users to a group

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.add_users_to_group('group_id', ['user_id_1', 'user_id_2'])
        True

        Parameters
        ----------
        group_id : str
            The group to add the users to
        user_ids : list of str
            The IDs of the users to add
        port : int (default 8086)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        success : bool
            Returns True if operation was successful
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _users_to_group(
            url,
            self.headers,
            group_id,
            user_ids,
            True
        )

    def remove_users_from_group(self, group_id: str, user_ids: list, port: int = 8086, use_port: bool = None) -> bool:
        """
        Remove users from a group

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.remove_users_from_group('group_id', ['user_id_1', 'user_id_2'])
        True

        Parameters
        ----------
        group_id : str
            The ID of the group
        user_ids : list of str
            The IDs of the users to remove
        port : int (default = 8086)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        success : bool
            Returns True if successful
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _users_to_group(
            url,
            self.headers,
            group_id,
            user_ids,
            False
        )

    def list_users(self, max_count: int = 100, as_df: bool = True, port: int = 8080, use_port: bool = None) -> Union[pd.DataFrame, dict]:
        """
        List all users

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.list_users()
        *DataFrame with results*

        Parameters
        ----------
        max_count : int (default 100)
            The maximum number of users to return
        as_df : bool (default True)
            Whether to return the data as a Pandas DataFrame
        port : int (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        users : pandas DataFrame or dictionary
            The response from the API

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_users(
            url,
            self.headers,
            max_count,
            as_df
        )

    def list_groups(self, max_count: int = 100, as_df: bool = True, port: int = 8083, use_port: bool = None) -> Union[pd.DataFrame, dict]:
        """
        List all groups

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.list_groups()
        *DataFrame with results*

        Parameters
        ----------
        max_count : int (default 100)
            The maximum number of groups to return
        as_df : bool (default True)
            Whether to return the result as a pandas DataFrame
        port : int (default 8083)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        groups : pandas DataFrame or dictionary
            The response from the API

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_groups(
            url,
            self.headers,
            max_count,
            as_df
        )

    def list_group_users(self, group_id: str, as_df: bool = True, port: int = 8083, use_port: bool = None) -> Union[pd.DataFrame, dict]:
        """
        List users in a group

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.list_group_users('group_id')
        *DataFrame with results*

        Parameters
        ----------
        group_id : str
            The ID for the group
        as_df : bool (default True)
            Whether to return the response as a pandas DataFrame
        port : int (default 8083)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        users : pandas DataFrame or dictionary
            The response from the API

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_group_users(
            url,
            self.headers,
            as_df,
            group_id
        )

    def list_roles(self, as_df: bool = True, port: int = 8086, use_port: bool = None) -> Union[pd.DataFrame, dict]:
        """
        List the roles available in the platform

        Example usage:

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.list_roles()
        *DataFrame with results*

        Parameters
        ----------
        as_df : bool (default True)
            Whether to return the results as a pandas DataFrame
        port : int (default 8086)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        roles : pandas DataFrame or dict
            The roles
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_roles(
            url,
            self.headers,
            as_df
        )

    # Metrics

    def list_user_usage_metrics(self, user_id: str, period: str = 'hourly', as_df: bool = True, port: int = 8080, use_port: bool = None) -> Union[dict, pd.DataFrame]:
        """
        Get usage metrics for a user

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.get_user_usage_metrics('user_id')
        *DataFrame with results*

        Parameters
        ----------
        user_id : str
            The ID of the user
        period : str (default 'hourly')
            The period to group metrics into
        as_df : bool (default True)
            Whether to return results as a pandas DataFrame
        port : int (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        results : pandas DataFrame or dict
            The results from the platform
        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_user_usage_metrics(
            url,
            self.headers,
            user_id,
            period,
            as_df
        )

    def list_model_usage_metrics(self, model_id: str, period: str = 'hourly', as_df: bool = True, port: int = 8080, use_port: bool = None) -> Union[dict, pd.DataFrame]:
        """
        Get usage metrics for a model

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.get_model_usage_metrics('model_id')
        *DataFrame with results*

        Parameters
        ----------
        model_id : str
            The ID of the model
        period : str (default 'hourly')
            The period to group metrics into
        as_df : bool (default True)
            Whether to return results as a pandas DataFrame
        port : int (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value


        Returns
        -------
        results : pandas DataFrame or dict
            The results from the platform

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _list_model_usage_metrics(
            url,
            self.headers,
            model_id,
            period,
            as_df
        )

    # Additional utilities

    def get_user_id_by_name(self, name: str, port: int = 8080, use_port: bool = None) -> str:
        """
        Get a user's ID from their display name

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.get_user_id_by_name('User Name')
        *user_id*

        Parameters
        ----------
        name : str
            The display name of the user
        port : int (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        id : str
            The ID of the user

        """

        users = self.list_users(port=port, use_port=use_port)
        this_user = users[users.displayName == name]

        if this_user.shape[0] == 0:
            raise ValueError('No user of that name appears to exist')

        return this_user.id.iloc[0]

    def get_model_id_by_name(self, model_name: str, port: int = 8080, use_port: bool = None) -> str:
        """
        Retrieve a model's ID using the name of the model

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.get_model_id_by_name('my_awesome_model')
        *model_id*

        Parameters
        ----------
        model_name : str
            The name of the model
        port : int (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value


        Returns
        -------
        model_id : str
            The model's ID

        """

        models = self.list_models(port=port, use_port=use_port)
        this_model = models[models.name == model_name]

        if this_model.shape[0] == 0:
            raise ValueError('No model with that name appears to exist')

        return this_model.id.iloc[0]

    def get_group_id_by_name(self, group_name: str, port: int = 8083, use_port: bool = None) -> str:
        """
        Get the ID of a group by searching for its display name

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.get_group_id_by_name('Group Name')
        *group_id*

        Parameters
        ----------
        group_name : str
            The display name of the group
        port : int (default 8083)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        group_id : str
            The ID of the group
        """

        groups = self.list_groups(port=port, use_port=use_port)
        this_group = groups[groups.name == group_name]

        if this_group.shape[0] == 0:
            raise ValueError('No group with that name appears to exist')

        return this_group.id.iloc[0]

    def get_role_id_by_role_name(self, role_name: str, port: int = 8086, use_port: bool = None) -> str:
        """
        Get the ID of a role by searching for its display name

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.get_role_id_by_role_name('Role Name')
        *role_id*

        Parameters
        ----------
        role_name : str
            The name of the role
        port : int (default 8086)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        role_id : str
            The ID of the role
        """

        roles = self.list_roles(port=port, use_port=use_port)
        this_role = roles[roles.name == role_name]

        if this_role.shape[0] == 0:
            raise ValueError('No role with that name appears to exist')

        return this_role.id.iloc[0]

    def test_connection(self, port: int = 8080, use_port: bool = None) -> bool:
        """
        Test whether there is a healthy connection to the platform

        >>> import aisquared
        >>> client = aisquared.platform.AISquaredPlatformClient()
        >>> client.test_connection()
        True

        Parameters
        ----------
        port : int (default 8080)
            The API port for the call. This can be handled automatically by the platform ALB
        use_port : bool or None (default None)
            Whether to use port in URL formatting. If None, defaults to class value

        Returns
        -------
        success : bool
            True if connection was successful

        """

        if use_port is None:
            use_port = self.use_port

        url = self._format_url(self.base_url, port, use_port)

        return _test_connection(self.base_url)



from typing import Union
from aisquared.base import BaseObject


class QueryParameterHarvester(BaseObject):
    """
    Harvester for Query Parameters

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.harvesting.QueryParameterHarvester(
        'test_key',
        'test_url',
        'test_attribute'
    )
    >>> my_obj.to_dict()
    {'className': 'QueryParameterHarvester',
    'params': {'queryKeys': ['test_key'],
    'urlLocations': ['test_url'],
    'attributes': ['test_attribute']}}

    """

    def __init__(
        self,
        query_keys: Union[str, list],
        url_locations: Union[str, list],
        attributes: Union[str, list]
    ):
        """
        Parameters
        ----------
        query_keys : list of str
            Keys to use in the query
        url_locations : list of str
            URL Locations to query from
        attributes : list of str
            Attributes for the query
        """
        self.query_keys = query_keys
        self.url_locations = url_locations
        self.attributes = attributes

    @property
    def query_keys(self):
        return self._query_keys

    @query_keys.setter
    def query_keys(self, value):
        if isinstance(value, str):
            value = [value]

        if not isinstance(value, list) or not all([isinstance(v, str) for v in value]):
            raise ValueError('query_keys must be list of strings')

        self._query_keys = value

    @property
    def url_locations(self):
        return self._url_locations

    @url_locations.setter
    def url_locations(self, value):
        if isinstance(value, str):
            value = [value]

        if not isinstance(value, list) or not all([isinstance(v, str) for v in value]):
            raise ValueError('url_locations must be list of strings')

        self._url_locations = value

    @property
    def attributes(self):
        return self._attributes

    @attributes.setter
    def attributes(self, value):
        if isinstance(value, str):
            value = [value]

        if not isinstance(value, list) or not all([isinstance(v, str) for v in value]):
            raise ValueError('attributes must be list of strings')

        self._attributes = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'QueryParameterHarvester',
            'params': {
                'queryKeys': self.query_keys,
                'urlLocations': self.url_locations,
                'attributes': self.attributes
            }
        }



from .AISquaredAPIException import AISquaredAPIException
from .additional_utils import _check_results_length
from aisquared.base import ENDPOINTS
import pandas as pd
import requests


def _list_user_usage_metrics(
        url: str,
        headers: dict,
        user_id: str,
        period: str,
        as_df: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    List the usage of the platform by a user

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    user_id : str
        The ID of the user
    period : int
        The period to group metrics into (e.g. 'hourly')
    as_df : bool
        Whether to return the data as a Pandas DataFrame
    """

    url = f'{url}/{ENDPOINTS["usage_metrics"]}?period={period}&entityId={user_id}&entity=user&action=run'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    if as_df:
        df = pd.DataFrame(resp.json()['data']['plotXYData'])
        _check_results_length(df)
        return df

    return resp.json()


def _list_model_usage_metrics(
        url,
        headers,
        model_id,
        period,
        as_df
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    List the usage of a model

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    model_id : str
        The ID of the user
    period : int
        The period to group metrics into (e.g. 'hourly')
    as_df : bool
        Whether to return the data as a Pandas DataFrame
    """

    url = f'{url}/{ENDPOINTS["usage_metrics"]}?period={period}&entity=model&entityId={model_id}&action=run'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

        if not resp.ok:
            raise AISquaredAPIException(resp.json())

        if as_df:
            df = pd.DataFrame(resp.json()['data']['plotXYData'])
            _check_results_length(df)
            return df

        return resp.json()



from typing import Union
from aisquared.base import BADGES, WORD_LISTS, BaseObject


class WordRendering(BaseObject):
    """
    Object for rendering badges on individual words

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.WordRendering()
    >>> my_obj.to_dict()
    {'className': 'WordRendering',
    'params': {'wordList': 'input',
    'resultKey': None,
    'contentKey': None,
    'badgeShape': 'star',
    'badgeColor': 'blue',
    'classes': None,
    'thresholdKey': None,
    'thresholdValue': None
    'position': 'after'}}

    """

    def __init__(
            self,
            word_list: str = WORD_LISTS[0],
            result_key: str = None,
            content_key: str = None,
            badge_shape: str = BADGES[-2],
            badge_color: str = 'blue',
            classes: list = None,
            threshold_key: str = None,
            threshold_value: Union[int, float] = None,
            position: str = 'after'
    ):
        """
        Parameters
        ----------
        word_list : str (default 'input')
            How to identify words to render, must be one of 'input', 'result'
        results_key : str or None (default None)
            The result key to use to render, only used if word_list is 'result'
        content_key : str or None (default None)
            The key from the results to use in rendering
        badge_shape : str (default 'star')
            The badge shape to use
        badge_color : str (default 'blue')
            The badge color to use
        classes : None or list (default None)
            If provided, list of classes that will be rendered
        threshold_key : None or string (default None)
            If provided, the key to look for to threshold
        theshold_value : None or numeric (default None)
            If provided with threshold_key, the minimum value required to render
        position : str (default 'after')
            The position of the rendering widget, either 'before' or 'after'
        """
        super().__init__()
        self.word_list = word_list
        self.result_key = result_key
        self.content_key = content_key
        self.badge_shape = badge_shape
        self.badge_color = badge_color
        self.classes = classes
        self.threshold_key = threshold_key
        self.threshold_value = threshold_value
        self.position = position

    @property
    def word_list(self):
        return self._word_list

    @word_list.setter
    def word_list(self, value):
        if value not in WORD_LISTS:
            raise ValueError(
                f'word_list must be one of {WORD_LISTS}, got {value}')
        self._word_list = value

    @property
    def result_key(self):
        return self._result_key

    @result_key.setter
    def result_key(self, value):
        self._result_key = value

    @property
    def content_key(self):
        return self._content_key

    @content_key.setter
    def content_key(self, value):
        self._content_key = value

    @property
    def badge_shape(self):
        return self._badge_shape

    @badge_shape.setter
    def badge_shape(self, value):
        if value not in BADGES:
            raise ValueError(f'badge_shape must be one of {BADGES}')
        self._badge_shape = value

    @property
    def badge_color(self):
        return self._badge_color

    @badge_color.setter
    def badge_color(self, value):
        self._badge_color = value

    @property
    def classes(self):
        return self._classes

    @classes.setter
    def classes(self, value):
        self._classes = value

    @property
    def threshold_key(self):
        return self._threshold_key

    @threshold_key.setter
    def threshold_key(self, value):
        self._threshold_key = value

    @property
    def threshold_value(self):
        return self._threshold_value

    @threshold_value.setter
    def threshold_value(self, value):
        self._threshold_value = value

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        if not isinstance(value, str):
            raise TypeError('position must be string')
        if value not in ['before', 'after']:
            raise ValueError(
                f'position must be either "before" or "after", got {value}')
        self._position = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'WordRendering',
            'params': {
                'wordList': self.word_list,
                'resultKey': self.result_key,
                'contentKey': self.content_key,
                'badgeShape': self.badge_shape,
                'badgeColor': self.badge_color,
                'classes': self.classes,
                'thresholdKey': self.threshold_key,
                'thresholdValue': self.threshold_value,
                'position': self.position
            }
        }



"""
The aisquared.config.feedback subpackage contains objects for configuring feedback in aisquared models.
"""

from .SimpleFeedback import SimpleFeedback
from .BinaryFeedback import BinaryFeedback
from .MulticlassFeedback import MulticlassFeedback
from .RegressionFeedback import RegressionFeedback
from .ModelFeedback import ModelFeedback
from .QualitativeFeedback import QualitativeFeedback



"""NOT MEANT TO BE CALLED BY THE END USER - configuration parameters for the different endpoints in the platform"""

ENDPOINTS = {
    'login': 'api/v1/auth/login',
    'upload_model': 'upload/v1/models',
    'model': 'api/v1/models',
    'feedback': 'api/v1/feedback',
    'user': 'userservice/v1/user',
    'group': 'groupservice/v1/group',
    'group_membership': 'groupservice/v1/membership',
    'user_list': 'scim/v2/Users',
    'group_list': 'scim/v2/Groups',
    'roles': 'groupservice/v1/role',
    'usage_metrics': 'api/v1/usage-metrics',
    'health': 'api/v1/health'
}
